package require XOTcl
namespace import ::xotcl::*

package require xdobry::sql

package require textutil

package require inflector
package require enumerable
package require extended_metaclass

proc parse_options {options args} {
	upvar $options _options
	upvar $args _args
	foreach {key val} $_args {
		set _options($key) $val
	}
}

@ @File {
	description {
	ActiveRecord::Base is a abstract class. Concrete class is generated as follows.
	
		set classname [ActiveRecord::Base ActiveRecord::User]
		=> User class is created.
	
	Connection to the users table is prepared in the class constructor.
	Existing records are collected by "find_by_sql" or other wrappers.
	In order to add a new record, create an object by "new", edit its property, 
	and "save" it.
	The class holds all of the record objects generated.
	}
}

namespace eval ActiveRecord {
	set SELECT_ORDER {
		"select"    "select distinct"    "update"    "insert"    "replace"    "delete"
		"from"      "set"       "into"
		"where"                 "values"
		"order by"
		"limit"     "offset"
	}
	
	Class Base -superclass Class -parameter {
		{use_replace 0}
	}
	
	Base instproc init {args} {
		my instvar table columns connection
		set table [Inflector::tableize [self]]
		set connection [my connect]
		set columns [my columns]
		
		set singularClassMethods [namespace tail [self]]ClassMethods
		catch {
			[self] mixin $singularClassMethods
		}
	}
	
	Base instproc destroy {args} {
		my instvar connection
		catch {
			$connection disconnect
			$connection destroy
		}
		next
	}
	
	Base instproc connect {} {
		my instvar db_interface db_connect_info
		
		my connection_prepare
		set connection [$db_interface [self]_connection]
		if {[$connection connect [array get db_connect_info]]} {
			error "connection error code=$::errorInfo"
		}
		return $connection
	}
	
	Base instproc connection_prepare {} {
		# See each adaptor.
	}
	
	Base instproc execute_sql {query} {
		my instvar connection
		join [$connection queryList $query]
	}
	
	Base instproc columns {} {
		my instvar dbinfo column_info
		set columns {}
		foreach $dbinfo(column_attributes) [my execute_sql $dbinfo(table_info_query)] {
			lappend columns $field
			foreach attribute $dbinfo(column_attributes) {
				lappend column_info($field) [subst $$attribute]
			}
		}
		return $columns
	}
	
	Base instproc column_attribute {column attribute} {
		my instvar dbinfo column_info
		
		set index [lsearch $dbinfo(column_attributes) $attribute]
		return [lindex $column_info($column) $index]
	}
	
	Base instproc value_for_sql {field property} {
		upvar $property _property
		switch [my reference_type $field $_property converted] {
			quote     {
				set value "'[string trim $_property ']'"
			}
			convert {
				if {[string is space $_property]} {
					set value "NULL"
				} else {
					set value $converted
				}
			}
			no-quote  {
				if {[string is space $_property]} {
					set value "NULL"
				} else {
					set value $_property
				}
			}
			reference {
				set value $$property
			}
		}
		return $value
	}
	
	Base instproc reference_type {field value converted} {
		# see each adaptor
		# returns either "quote", "no-quote" or "reference"
	}
	
	Base instproc my_instances {} {
		enumerable::select "{string match \[\$x class\] [self]}" {x} [my info children]
	}
	
	Base instproc unknown {cmd args} {
		if       {[regsub {^find_all_by_} $cmd {} ret]} {
			eval my find_by $ret all   $args
		} elseif {[regsub {^find_by_}     $cmd {} ret]} {
			eval my find_by $ret first $args
		} elseif {[regsub {^find_all$}    $cmd {} ret]} {
			     my find_by $ret all
		} elseif {[regsub {^find_first$}  $cmd {} ret]} {
			     my find_by $ret first
		} else {
			next
		}
	}
	
	Base instproc create_sql {sql} {
		variable SELECT_ORDER
		upvar $sql _sql
		set query {}
		foreach key $SELECT_ORDER {
			if [info exists _sql($key)] {lappend query $key $_sql($key)}
		}
		join $query
	}
	
	Base instproc find_by_sql {sql} {
		upvar $sql _sql
		
		set fields [split [regsub {^(distinct|all) } $_sql(select) {}] ,]
		if {$fields == "*"} {set fields [my set columns]}
		
		set query [my create_sql _sql]
		set records {}
		foreach $fields [my execute_sql $query] {
			set record [[self] new -childof [self]]
			foreach field $fields {
				$record set property($field) [set $field]
			}
			$record define_association_methods
			lappend records $record
		}
		return $records
	}
	
	Base instproc find_by {fields limit args} {
		my instvar table
		if {$limit == "all"} {
		} else {
			set sql(limit) 1
		}
		array set sql [list \
			"select"    "*" \
			"from"      "$table" \
		]
		set fields [textutil::splitx $fields {_and_}]
		if {$fields != {}} {
			foreach $fields $args {
				set where {}
				foreach column $fields {
					lappend where "$column = [my value_for_sql $column $column]"
				}
				set sql(where) [join $where " and "]
			}
		}
		my find_by_sql sql
	}
	
	Base instproc inspect {} {
		set records [my info children]
		foreach record $records {
			puts -nonewline -
			foreach key [$record bare_fields] {
				puts "${key}  [$record set property($key)]"
			}
			puts {}
		}
		puts "[llength $records] [self] Instances"
		return $records
	}
	
	Base instproc destroy_objects {records} {
		foreach record $records {
			catch {$record destroy}
		}
	}
	
	Base instproc count {column {where {}}} {
		my instvar table
		array set sql [list \
			select count($column) \
			from $table \
		]
		if {$where != {}} {set sql(where) $where}
		my execute_sql [my create_sql sql]
	}
	
	Base instproc selectableItems {column} {
		my instvar table
		array set sql [list \
			"select distinct" $column \
			from $table \
		]
		my execute_sql [my create_sql sql]
	}
	
	Class InstanceMethods
	
	InstanceMethods instproc init {args} {
		my array set property {}
	}
	
	InstanceMethods instproc destroy {} {
		next
	}
	
	InstanceMethods instproc define_association_methods {} {
		my instvar property
		
		variable associations
		foreach association $associations {
			foreach {subject verb object} $association break
			set target [Inflector::singularize [my cset table]]
			
			set has_one    [list $target has_one    $object]
			set has_many   [list $target has_many   $object]
			set belongs_to [list $target belongs_to $object]
			
			if {[lsearch [list $has_one $has_many $belongs_to] [list $subject $verb $object]] >= 0} {
				switch $verb {
					has_one {
						[self] proc $object {} "
							ActiveRecord::[Inflector::classify $object] find_by_${target}_id $property(id)
						"
					}
					has_many {
						[self] proc $object {} "
							ActiveRecord::[Inflector::classify $object] find_all_by_${target}_id $property(id)
						"
					}
					belongs_to {
						if {[info exists property(${object}_id)]} {
							[self] proc $object {} "
								my instvar property
								ActiveRecord::[Inflector::classify $object] find_by_id \[set property(${object}_id)\]
							"
						} else {
							[self] proc $object {} {return {}}
						}
					}
				}
			}
		}
	}
	
	InstanceMethods instproc create_sql {sql} {
		upvar $sql _sql
		[my class] create_sql _sql
	}
	
	InstanceMethods instproc fields {} {
		set fields {}
		foreach field [my cset columns] {
			if {[lsearch [my array names property] $field] == -1} {continue}
			lappend fields [format {"%s"} $field]
		}
		return $fields
	}
	
	InstanceMethods instproc bare_fields {} {
		set fields {}
		foreach field [my cset columns] {
			if {[lsearch [my array names property] $field] == -1} {continue}
			lappend fields $field
		}
		return $fields
	}
	
	# -withid [true|false]  : include id field or not.
	# -expression [tcl|sql] : add quotes for strings, NULL for empty field.
	InstanceMethods instproc fields_and_values {{args {-withid true -expression sql}}} {
		parse_options options args
		set fields {}
		set values {}
		foreach bare_field [my bare_fields] {
			if {$bare_field == "id" && $options(-withid) == false} {continue}
			lappend fields [format {"%s"} $bare_field]
			lappend values [[my class] value_for_sql $bare_field [myvar property($bare_field)]]
		}
		return [list $fields $values]
	}
	
	InstanceMethods instproc save {} {
		my create_or_update
	}
	
	InstanceMethods instproc create_or_update {} {
		my instvar property
		if {[array names property id] == {} || $property(id) == {}} {
			set new_record 1
		} else {
			set id $property(id)
			set count [[my class] count id "id = $id"]
			set new_record [expr $count ? 0 : 1]
		}
		
		set result [expr $new_record ? {[my create_without_id]} : {[my update_or_replace]}]
	}

	InstanceMethods instproc create_without_id {} {
		my create false
	}

	InstanceMethods instproc create_with_id {} {
		my create true
	}

	InstanceMethods instproc create {{withid false}} {
		foreach {fields values} [my fields_and_values -withid $withid] break
		array set sql [list \
			"insert" "" \
			"into"   "[my cset table] ([join $fields ,])" \
			"values" "([join $values ,])" \
		]
		[my class] execute_sql [my create_sql sql]
		if {!$withid} {
			set newid [[my class] execute_sql [my cset dbinfo(max_id_query)]]
			my set property(id) $newid
		}
		my define_association_methods
	}

	InstanceMethods instproc update_or_replace {} {
		set result [expr [my cset use_replace] ? {[my replace]} : {[my update]}]
	}

	InstanceMethods instproc replace {} {
		foreach {fields values} [my fields_and_values -withid true] break
		array set sql [list \
			"replace" "" \
			"into"   "[my cset table] ([join $fields ,])" \
			"values" "([join $values ,])" \
		]
		[my class] execute_sql [my create_sql sql]
		my define_association_methods
	}

	InstanceMethods instproc update {} {
		my instvar property
		foreach {fields values} [my fields_and_values -withid false] break
		[my class] execute_sql [my cset dbinfo(begin)]
		for {set i 0} {$i < [llength $fields]} {incr i} {
			array set sql [list \
				"update" "[my cset table]" \
				"set"    "[lindex $fields $i] = [lindex $values $i]" \
				"where"  "id = $property(id)" \
			]
			[my class] execute_sql [my create_sql sql]
		}
		[my class] execute_sql [my cset dbinfo(commit)]
		my define_association_methods
	}

	InstanceMethods instproc delete {} {
		my instvar property
		[my class] execute_sql [my cset dbinfo(begin)]
		if {[info exists property(id)]} {
			array set sql [list \
				"delete" "" \
				"from"   "[my cset table]" \
				"where"  "id = $property(id)" \
			]
			[my class] execute_sql [my create_sql sql]
		}
		[my class] execute_sql [my cset dbinfo(commit)]
		my destroy
	}

	InstanceMethods instproc inspect {} {
		my instvar property
		append data "[self]\n-"
		foreach key [my bare_fields] {
			append data "${key} $property($key)\n"
		}
		return $data
	}

}

package provide activerecord 0.2
